var playListUpdate = false;
var player = null;
var tracks = null;
var activeTimeout = 0;
var activeTrackItem = null;

function init()
{
	openPlayList();
	player = Components.classes["@XulMusic.org;Player;1"].getService(Components.interfaces.xMPlayer);
	tracks = new Array();
	playListUpdate = false;

	var trackEvents = {
		onTrackFinished: function(){

			// because we need to give the underlying implementation 
			// a chance to clean up we put the logic here in a timeout
			setTimeout( function(){
				dump( "calling stop\n" );
				player.stop();
				dump( "calling nextTrack\n" );
				nextTrack();
			}, 1000 );
		}
	};

	player.setTrackListener( trackEvents );

}
function nextTrack()
{
	var list = document.getElementById( "play-list" );
	var items = list.getElementsByTagName( "listitem" );
	for( var i = 0; i < items.length; ++i ){
		var item = items[i];
		if( item == activeTrackItem ){
			var next = i + 1;
			if( next >= items.length ){
				next = 0; // loop
			}
			dump( "found active track playing next track\n" );
			list.selectedItem = items[next];
			playSelected();
			break;
		}
	}
}

function cleanup()
{
	savePlayList();
}

function addTrack()
{
	var fpicker = Components.classes["@mozilla.org/filepicker;1"].
								createInstance(Components.interfaces.nsIFilePicker);

	fpicker.init( window, "Restore an existing Backup",
								Components.interfaces.nsIFilePicker.modeOpen );
	fpicker.appendFilter( "MP3", "*.mp3" );
	fpicker.appendFilter( "Ogg Vorbis", "*.ogg" );

	var ret = fpicker.show();
	if( ret == Components.interfaces.nsIFilePicker.returnOK ||
			ret == Components.interfaces.nsIFilePicker.returnReplace ){
		//fpicker.fileURL.spec
		// TODO: we should create the track here and determine the tracks
		// title and duration
		var list = document.getElementById( "play-list" );
		//list.appendItem( fpicker.file.leafName, fpicker.file.path );

		var listitem = document.createElement( "listitem" );
		var cell = document.createElement( "listcell" );
		cell.setAttribute( "label", fpicker.file.leafName );
		cell.setAttribute( "value", fpicker.file.path );
		listitem.appendChild( cell );
		list.appendChild( listitem );

		playListUpdate = true;
	}
}

function removeSelectedTrack()
{
	var list = document.getElementById( "play-list" );
	if( list.selectedItem == null ){
		dump( "no item selected\n" );
		return;
	}
	list.removeItemAt( list.selectedIndex );
	playListUpdate = true;
}

function showHidePlayList()
{
	// TODO: create artwork for the playlist closed state
	// toggle playlist closed state
	var playList = document.getElementById( "play-list-shell" );
	var control = document.getElementById( "playlist-show-hide" );
	var playlistHeight = document.getElementById( "controls" ).getAttribute( "playlist-height" );
	var xulWindow = document.getElementById( "xul-music" );

	if( playList.getAttribute("hidden") == "true" ){
		playList.setAttribute( "hidden", "false" );
		xulWindow.setAttribute( "height", playlistHeight );
		window.resizeTo( window.innerWidth, playlistHeight );
	}
	else{
		document.getElementById( "controls" ).setAttribute( "playlist-height", 
																					xulWindow.getAttribute("height") );
	// XXX: get the computed height of the controls box instead of
	// hard coding the height.
		playList.setAttribute( "hidden", "true" );
		xulWindow.setAttribute( "height", "85" );
		window.resizeTo( window.innerWidth, 85 );
	}
}

function savePlayList()
{
	if( !playListUpdate )
		return;

	dump( "writing playlist\n" );

	//var doc = document.implementation.createDocument("","",null);
	var list = document.getElementById( "play-list" );
	var items = list.getElementsByTagName( "listitem" );
	var data = "<?xml version=\"1.0\"?>\n";

	data += "<items>\n";
	for( var i = 0; i < items.length; ++i ){
		var cells = items[i].getElementsByTagName( "listcell" );
		data += "\t<item>\n";
		for( var j = 0; j < cells.length; ++j ){
			data += "\t\t<cell label='" + cells[j].getAttribute( "label" ) + 
											"' value='" + cells[j].getAttribute( "value" ) + "'/>\n";
		}
		data += "\t</item>\n";
		//data += "\t<item label='" + items[i].label + "' value='" + items[i].value + "'/>\n";
	}
	data += "</items>\n";

	var file = getXmlPlayListFile();
	dump( "writing file: " + file.path + "\n" );
	// file is nsIFile, data is a string
	var foStream = Components.classes["@mozilla.org/network/file-output-stream;1"]
		.createInstance(Components.interfaces.nsIFileOutputStream);
	//var serializer = new XMLSerializer();
	//var data = serializer.serializeToString( doc );
	dump( "writing data\n" + data + "\n" );
	// use 0x02 | 0x10 to open file for appending.
	foStream.init(file, 0x02 | 0x08 | 0x20, 0664, 0); // write, create, truncate
	foStream.write(data, data.length);
	foStream.close();
}
function getXmlPlayListFile()
{
	// get profile directory
	var file = Components.classes["@mozilla.org/file/directory_service;1"].
												getService(Components.interfaces.nsIProperties).
												get("ProfD", Components.interfaces.nsIFile);
	
	file.append( "xmplaylist.xml" );
	if( !file.exists() ){
		file.create( Components.interfaces.nsIFile.NORMAL_FILE_TYPE, 0664 );
	}
	return file;
}
function getXmlPlayListFileURL()
{
	var ios = Components.classes["@mozilla.org/network/io-service;1"]
											.getService(Components.interfaces.nsIIOService);
	return ios.newFileURI(getXmlPlayListFile());
}

function openPlayList()
{
	var fileURL = getXmlPlayListFileURL().spec;
	var xmlhttp = new XMLHttpRequest();

	xmlhttp.onreadystatechange = function(){
		// status == 0 since we're reading a file from the 
		// profile folder
		if( xmlhttp.readyState == 4 && xmlhttp.status == 0 ){
			var list = document.getElementById( "play-list" );
			var items = xmlhttp.responseXML.getElementsByTagName( "item" );
			for( var i = 0; i < items.length; ++i ){
				var cells = items[i].getElementsByTagName( "cell" );
				var listitem = document.createElement( "listitem" );
				for( var j = 0; j < cells.length; ++j ){
					var cell = document.createElement( "listcell" );
					cell.setAttribute( "label", cells[j].getAttribute("label") );
					cell.setAttribute( "value", cells[j].getAttribute("value") );
					listitem.appendChild( cell );
				}
				list.appendChild( listitem );
			}
		}
	};

	dump( fileURL + "\n" );
	xmlhttp.open( "GET", fileURL, true );
	xmlhttp.overrideMimeType("text/xml");
	xmlhttp.send(null);
}

function getCurrentPlayListItem()
{
	// determine the selected track
	var list = document.getElementById( "play-list" );
	var item = null;
	if( list.selectedItem == null ){
		var items = list.getElementsByTagName( "listitem" );
		if( items.length > 0 ){
			item = items[0];
			list.selectedItem = item;
		}
	}
	else
		item = list.selectedItem;

	// if no item is currently selected return
	if( !item )
		return null;

	return item;//.getElementsByTagName( "listcell" )[0];
}

function playFile( filePath, item )
{
	var track = tracks[filePath];
	var updateDuration = false;
	if( !track ){
		// create the new track and store it in our track list (aka playlist)
		track = player.newTrack( newLocalFile( filePath ) );
		tracks[filePath] = track;
		updateDuration = true;
	}

	var tracker = document.getElementById( "curtrack-position" );
	tracker.setAttribute( "curpos", "0" );

	// set the current track
	player.track = track;
	// start playing
	player.play();

	// store the activeItem so that we can update it later
	activeTrackItem = item;

	activeTimeout = setTimeout( progressUpdate, 1000 );

	/*if( updateDuration && activeTrackItem.getElementsByTagName( "listcell" ).length == 1 ){
		updateCurrentTrackDuration();
	}*/
}

function updateCurrentTrackDuration()
{
	var track = player.track;
	if( track ){
		var duration = track.getDuration();
		var cells = activeTrackItem.getElementsByTagName( "listcell" );
		if( cells.length > 1 ){
			var cell = cells[1];
			cell.setAttribute( "value", duration );
			cell.setAttribute( "label", getTimeString(duration) );
		}
		else{
			var cell = document.createElement( "listcell" );
			dump( "track.duration: " + track.getDuration() + "\n" );
			cell.setAttribute( "label", getTimeString(duration) );
			cell.setAttribute( "value", duration );
			activeTrackItem.appendChild( cell );
		}
		playListUpdate = true;
	}
}

function getCurrentTrackDuration()
{
	var cells = activeTrackItem.getElementsByTagName( "listcell" );
	if( cells.length > 1 ){
		var cell = cells[1];
		return cell.getAttribute( "value" );
	}
	
	return 0;
}

function getTimeString( seconds )
{
	var min = Math.round( seconds / 60 );
	var sec = seconds % 60;

	return min + ":" + sec;
}

function progressUpdate()
{
	var track = player.track;
	var state = track.getState();
	var time = track.getPosition();
	var duration = track.getDuration();
	var tracker = document.getElementById( "curtrack-position" );
	var percent = Math.round( (time / duration) * 100.0 );
	var trackTime = document.getElementById( "track-time" );
	
	var ctime = getTimeString( time );
	var dtime = getTimeString( duration );

	tracker.setAttribute( "curpos", percent );

	trackTime.setAttribute("value", ctime);
	//dump( "progress: " + percent + "% --> " + ctime + " of " + dtime + "\n" );

	// update the track duration
	if( getCurrentTrackDuration() != duration ){
		updateCurrentTrackDuration();
	}

	// update the track name
	if( track.getTitle() != "" && track.getArtist() != "" ){
		var cells = activeTrackItem.getElementsByTagName( "listcell" );
		var cell = cells[0];
		var label = track.getArtist() + " - " + track.getTitle();
		if( label != cell.getAttribute("label") ){
			cell.setAttribute( "label", label );
			playListUpdate = true;
		}
	}

	if( state == Components.interfaces.xMTrack.PLAYING )
		activeTimeout = setTimeout( progressUpdate, 1000 );
}

function play()
{
	var button = document.getElementById( "track-play-pause" );
	if( button.getAttribute( "mode" ) == "play" ){

		var item = getCurrentPlayListItem();
		var cells = item.getElementsByTagName( "listcell" );
		var filePath = cells[0].getAttribute("value");
		if( !filePath )
			return;

		document.getElementById( "track-title" ).setAttribute( "value", cells[0].getAttribute("label") );

		playFile( filePath, item );

		button.setAttribute( "mode", "pause" );
	}
	else{
		if( activeTimeout ){
			window.clearTimeout( activeTimeout );
			activeTimeout = 0;
		}
		player.pause();
		button.setAttribute( "mode", "play" );
	}
}

function playSelected()
{
	var button = document.getElementById( "track-play-pause" );

	var item = getCurrentPlayListItem();
	var cells = item.getElementsByTagName( "listcell" );
	var filePath = cells[0].getAttribute("value");
	if( !filePath )
		return;

	document.getElementById( "track-title" ).setAttribute( "value", cells[0].getAttribute("label") );

	if( activeTimeout ){
		window.clearTimeout( activeTimeout );
		activeTimeout = 0;
	}
	// stop whatever might be playing
	player.stop();

	playFile( filePath, item );

	// set the button state
	button.setAttribute( "mode", "pause" );

}

function startPositionChange( scroll )
{
	dump( "start change\n" );
	scroll.setAttribute("changing","true");
	if( activeTimeout ){
		window.clearTimeout( activeTimeout );
		activeTimeout = 0;
	}
}
function endPositionChange( scroll )
{
	dump( "end change\n" );
	scroll.setAttribute("changing","false");

	dump( "get track\n" );
	var track = player.track;
	dump( "got track\n" );
	if( track ){
		dump( "get track state\n" );
		var state = track.getState();

		if( state == Components.interfaces.xMTrack.PLAYING || 
				state == Components.interfaces.xMTrack.PAUSED ){
			var pos = scroll.getAttribute("curpos") / 100;
			var duration = track.getDuration();
			dump( "set position to: " + (duration * pos) + "\n" );
			track.setPosition( duration * pos );
		}
	}
	activeTimeout = setTimeout( progressUpdate, 1000 );
}

function updatePosition( scroll )
{
	if( scroll.getAttribute("changing") == "true" ){
		var pos = scroll.getAttribute("curpos");
		dump( pos + "\n" );
		//player.track.setPosition( );
	}
}

function stop()
{
	if( activeTimeout ){
		window.clearTimeout( activeTimeout );
		activeTimeout = 0;
	}
	var tracker = document.getElementById( "curtrack-position" );
	tracker.setAttribute( "curpos", "0" );
	player.stop();
	var button = document.getElementById( "track-play-pause" );
	button.setAttribute( "mode", "play" );
}

// create a new nsILocalFile
function newLocalFile( path )
{
	var file = Components.classes["@mozilla.org/file/local;1"].createInstance(
	                          Components.interfaces.nsILocalFile);
	file.initWithPath( path );
	return file;
}
