/**
 * Stars!
 * コマンド関連クラス群
 *
 * $Id: stars_cmd.js,v 1.16 2006/02/12 11:07:30 shio Exp $
 */

ECN_UP_COST = 50;
COLONY_COST = 20;
POP_MOVE_COST = 5;
BASE_COST = 20;
FLEET_BUILD_COST = 30;
FLEET_MENT_COST = 2;
SETTLER_BUILD_COST = 10;
SETTLER_MENT_COST = 2;

SETTLER_MOVE_COST = 0;
SETTLER_JOIN_COST = 0;
SETTLER_COLONY_COST = 20;

CMD_ID_NOTHING = 0;
CMD_ID_ECN_UP = 1;
CMD_ID_COLONY = 2;
CMD_ID_POP_MOVE = 3;
CMD_ID_BASE = 4;
CMD_ID_FLEET = 5;
CMD_ID_SETTLER = 6;

CMD_ID_SETTLER_MOVE = 101;
CMD_ID_SETTLER_JOIN = 102;
CMD_ID_SETTLER_COLONY = 103;

function cmdExecute() {
    if (this.planet) {
        if (this.empire.id != this.planet.ctl.id) {
            return;
        }
    }
    if (this.empire.trs < this.cost() ) {
        return;
     }
     // コマンドを実行する前と実行した後ではコストが変わる場合がある！
    var cost = this.cost();        // だから実行前にコストを保存
    if (this.localExecute()) {
        this.empire.trs -= cost;
     }
}

// 経済開発
function CmdBuildEcn(empire, planet) {
    this.id = CMD_ID_ECN_UP;
    this.name = "経済開発";
    this.empire = empire;
    this.planet = planet;
    this.cost = function() {
        return ECN_UP_COST;
 //       return (this.planet.ecn + 1)  * (this.planet.ecn + 1) * 3;
    }
    this.localExecute = function() {
        var p = this.planet;
        if (p.ecn >= 9) {
            return false;
        }
        if (p.dpl == 0) {
            return false;
        }
        if (p.ecn >=  p.dpl * 2) {
            return false;
        }
        if (p.ecn >= p.shogyoKenTotalDpl()) {
            return false;
        }
        p.ecn += 1;
printConsole(p.ctl.id, "経済力増加 : " + p.name + "(" + p.ecn + ")");
        return true;
    }

    this.cmdExecute = cmdExecute;

    // 実行可能かどうかを考えて、設定する。
    this.think = function(ai, p) {
        if (p.ecn < 9) {
            if ( (p.ecn < p.shogyoKenTotalDpl()) && (p.ecn < (p.dpl * 2) )) {
              if (ai.empire.tmp_trs >= this.cost()) {
                    return true;
                }
            }
        }
       return false;
    }

    this.canBuild = this.think;

    this.canBuildList = function(ai, p) {
        if (this.canBuild(ai, p)) {
            return new Array(this);
        }
    }


    this.getFullName = function() {
        return this.name;
    }

    this.makeKey = function() {
        return this.id;
    }
}

// 防衛力強化
function CmdBuildDfc(empire, planet) {
    this.id = CMD_ID_BASE;
    this.name = "防衛力強化";
    this.empire = empire;
    this.planet = planet;

    this.cost = function() {
        return BASE_COST;
    }

    this.localExecute = function() {
        var p = this.planet;
        if ( ( (p.dfc + 1) ) <= parseInt(p.dpl / 2 + 0.5) )  {
            p.dfc += 1;
            printConsole(p.ctl.id, "防衛力強化 : " + p.name + "(" + p.dfc + ")");
            return true;
        }
    }

    this.cmdExecute = cmdExecute;


    this.think = function(ai, p) {
        return (ai.checkBuildDfc(p));
    }

    this.canBuild = this.think;
    this.getFullName = function() {
        return this.name;
    }

    this.canBuildList = function(ai, p) {
        if (this.canBuild(ai, p)) {
            return new Array(this);
        }
    }

    this.makeKey = function() {
        return this.id;
    }
}

// 艦隊建設
function CmdBuildFleet(empire, planet) {
    this.id = CMD_ID_FLEET;
    this.name = "艦隊建設";
    this.empire = empire;
    this.planet = planet;

    this.cost = function() {
        return FLEET_BUILD_COST;
    }

    this.localExecute = function() {
        if (!this.canBuild(this.empire.ai, this.planet)) {
            return false;
        }
        var u = this.empire.getKaisanFleetByPlanetId(this.planet.id);
        if (u) {
            u.mlt = this.empire.mlt;
            u.nvg = this.empire.nvg;
            this.empire.removeKaisanFleet(u);
            printConsole(this.empire.id, u.getName() + "置換 : " + this.planet.name);
        } else {
            this.empire.buildFleet(this.planet);
            u = this.empire.getFleetByPlanetId(this.planet.id)
            printConsole(this.empire.id, u.getName() + "新設 : " + this.planet.name);
        }
        return true;
    }

    this.cmdExecute = cmdExecute;

    this.canBuild = function(ai, p) {
        if  (p.ecn < 5) {
            return false;
        }
        var cmd = app.getCmdByPlanetId(p.id);
        var cost = (cmd) ? cmd.cost : 0;
        if ( (this.empire.tmp_trs + cost) < FLEET_BUILD_COST) {
            return false;
        }
        if (this.empire.getFleetByPlanetId(p.id)) {
            if (!this.empire.getKaisanFleetByPlanetId(p.id)) {
                return false;
            }
        }
        
        return true;
    }
    this.canBuildList = function(ai, p) {
        if (this.canBuild(ai, p)) {
            return new Array(this);
        }
    }

    this.think = function(ai, p) {
        return (ai.checkBuildFleet(p) );
    }
    this.getFullName = function() {
        return this.name;
    }

    this.makeKey = function() {
        return this.id;
    }
}

// 植民処理
function CmdBuildColony(empire, from_p, to_p) {
    this.id = CMD_ID_COLONY;
    this.name = "植民";
    this.empire = empire;
    this.to_p = to_p;
    this.planet = from_p;

    this.cost = function() {
            return COLONY_COST;
    }

    this.localExecute = function() {
        if (! this.to_p) {
printConsole(this.empire.id, this.planet.name + ":植民失敗:植民先未設定");
            return false;
        }
        if (this.planet.dpl <= 1) {
printConsole(this.empire.id, this.planet.name + ":植民失敗:人口が１以下");
            return false;
        }

       if (this.planet.ctl.id != this.empire.id) {
printConsole(this.empire.id, this.planet.name + ":植民失敗:支配者が違う");
            return false;
        }

       if (this.to_p.ctl) {
printConsole(this.empire.id, this.planet.name + ":植民失敗:既に他の勢力に植民されている");
            return false;
        }

        this.planet.dpl -= 1;
        this.to_p.dpl = 1;
        this.to_p.changeCtl(this.empire);
printConsole(this.empire.id, this.empire.name + "は" + this.to_p.name + "にコロニーを築きました");
        return true;
    }

    this.cmdExecute = cmdExecute;

    this.think = function(ai, p) {
        if (! this.canBuild(ai, p) ) {
            return false;
        }
        return ai.checkBuildColony(p);
    }

    this.canBuild = function(ai, p) {
        return (this.canBuildList(ai, p).length > 0);
    }

    this.canBuildList = function(ai, p) {
        var result = new Array();
        if ( (ai.empire.tmp_trs >= COLONY_COST)
                && (p.dpl > 1) && ( (p.dpl - 1) * 2  >= p.ecn) ) {
            var undplList = ai.findUndplListFromLoc(p.loc, ai.empire.fnd);
            var list = ai.findColonizableList(undplList, ai.empire.nvg);
            for (var i = 0; i < list.length; i++) {
                var loc = list[i].loc;
                if (app.uchuju.isUchujuLoc(loc) ) {
                    continue;
                }
                if (! this.checkDupulicate(app.starList[loc])) {
                    this.to_p = app.starList[loc];
                    result.push(clone(this));
                }
            }
        }
        return result;
    }


    this.checkDupulicate = function(to_p) {
        for (var i = 0; i < app.commandList.length; i++) {
            var cmd = app.commandList[i];
            if (cmd.planet.id == this.planet.id) {
                continue;
            }
            if (cmd.id == CMD_ID_POP_MOVE) {
                if (cmd.from_p) {
                    if (cmd.from_p.id == this.planet.id) {
                        return true;
                    }
                }
            }
            if (cmd.id != this.id) {
                continue;
            }
            if (cmd.empire.id != this.empire.id) {
                continue;
            }
            if (cmd.to_p.id == to_p.id) {
                return true;
            }
        }
        return false;
    }

    this.getFullName = function() {
        return this.name + "(No." + this.to_p.id + ")";
    }

    this.makeKey = function() {
        return this.id + "-" + this.to_p.id;
    }
}

// 移民処理
function CmdImin(empire, to_p, from_p) {
    this.id = CMD_ID_POP_MOVE;
    this.name = "移民";
    this.empire = empire;
    this.from_p = from_p;
    this.planet = to_p;

    this.cost = function() {
            return POP_MOVE_COST;
    }

    this.localExecute = function() {
        if (! this.from_p) {
            return false;
        }
        if (this.from_p.ctl.id != this.empire.id) {
            return false;
        }

        if (this.planet.ctl) {
        } else {
            return false;
        }

        if (this.planet.ctl.id != this.empire.id) {
            return false;
        }

        if (this.from_p.dpl <= 1) {
            return false;
        }
        if (this.planet.dpl == 0) {
            return false;
        }
        if (this.planet.dpl == 9) {
            return false;
        }

        printConsole(this.empire.id, this.name + "：" + this.from_p.name
                + "->" + this.planet.name);
        this.from_p.dpl -= 1;
        this.planet.dpl += 1;

        return true;
    }

    this.cmdExecute = cmdExecute;

    this.think = function(ai, p) {
        if ( (p.dpl < 9) && (ai.empire.tmp_trs >= POP_MOVE_COST) ) {
            if (p.ecnLevel() >= 1) {
                var list = p.shogyoKenList();
                list = list.randomize();
                for (var j = 0; j < list.length; j++) {
                    var fp = list[j];
                    if (fp.dpl <= 1) {
                        continue;
                    }
                    if (fp.ecnLevel(-1) < p.ecnLevel() ) {
                        if (this.checkDuplicate(fp) ) {
                            continue;
                        }
                        this.from_p = fp;
                        this.planet = p;
                        return true;
                    } else if (fp.ecnLevel(-1) == p.ecnLevel() ) {
                        if (fp.dpl > p.dpl) {
                            if (this.checkDuplicate(fp) ) {
                                continue;
                            }
                            this.from_p = fp;
                            this.planet = p;
                            return true;
                        }
                    }
                }
            }
        }
        return false;
    }

    this.checkDuplicate = function(from_p) {
        for (var i = 0; i < app.commandList.length; i++) {
            var cmd = app.commandList[i];
            if (cmd.planet.id == this.planet.id) {
                continue;
            }
            if (cmd.id == CMD_ID_POP_MOVE) {
                if (cmd.from_p && from_p) {
                    if (cmd.from_p.id == from_p.id) {
                        return true;
                    }
                }
            }
        }
        return false;
    }


    this.canBuild = this.think;

    this.canBuildList = function(ai, p) {
        var result = new Array();

        if ( (p.dpl < 9) && (ai.empire.tmp_trs >= POP_MOVE_COST) ) {
            if (p.ecnLevel() >= 1) {
             var list = p.shogyoKenList();
             list = list.randomize();
             for (var j = 0; j < list.length; j++) {
                var fp = list[j];
                if (fp.dpl <= 1) {
                    continue;
                }
                if (fp.ecnLevel(-1) < p.ecnLevel() ) {
                        if (this.checkDuplicate(fp) ) {
                            continue;
                        }
                        this.from_p = fp;
                        this.planet = p;
                        result.push(clone(this));
                    } else if (fp.ecnLevel(-1) == p.ecnLevel() ) {
                        if (fp.dpl > p.dpl) {
                            if (this.checkDuplicate(fp) ) {
                                continue;
                            }
                            this.from_p = fp;
                            this.planet = p;
                            result.push(clone(this));
                        }
                    }
                }
            }
        }
        return result;
    }

    this.getFullName = function() {
        return this.name + "(" + this.from_p.name + ")";
    }

    this.makeKey = function() {
        return this.id + "-" + this.from_p.id;
    }
}

// 移民船団
function CmdBuildSettler(empire, planet) {
    this.id = CMD_ID_SETTLER;
    this.name = "移民船団建設";
    this.empire = empire;
    this.planet = planet;

    this.cost = function() {
        return SETTLER_BUILD_COST;
    }

    this.localExecute = function() {
        if (! this.planet.ctl) {
            return false;
        }
        if (this.planet.ctl.id != this.empire.id) {
            return false;
        }
        if (this.planet.dpl < 8) {
            return false;
        }
        this.planet.dpl--;
        this.empire.addSettler(planet.loc);
        printConsole(this.empire.id, "移民船団建設 : " + this.planet.name);
        return true;
    }

    this.cmdExecute = cmdExecute;

    this.canBuild = function(ai, p) {
        if  (p.dpl < 8) {
            return false;
        }
        if (this.empire.tmp_trs < SETTLER_BUILD_COST) {
            return false;
        }
        return true;
    }
    this.canBuildList = function(ai, p) {
        if (this.canBuild(ai, p)) {
            return new Array(this);
        }
    }

    this.think = function(ai, p) {
        if (this.empire.settlerList.length > 3) {
            return false;
        }
        if (this.empire.tmp_trs < 100) {
            return false;
        }
        return (ai.checkBuildSettler(p) );
    }
    this.getFullName = function() {
        return this.name;
    }

    this.makeKey = function() {
        return this.id;
    }
}

// コマンド未選択
function CmdNothing(empire, p) {
    this.empire = empire;
    this.planet = p;
    this.id = CMD_ID_NOTHING;
    this.name = "なし";
    this.cost = function() {
        return 0;
    }
    this.localExecute = function() {
        return true;
    }

    this.think = function() {
        return false;
    }

    this.canBuild = this.think;
    this.canBuildList = function() {
        return null;
    }

    this.getFullName = function() {
        return this.name;
    }

    this.makeKey = function() {
        return this.id;
    }
}

// 移民船団の移動
function CmdSettlerMove(empire, fromloc, toloc) {
    this.empire = empire;
    this.fromloc = fromloc;
    this.toloc = (toloc) ? toloc : null;
    this.id = CMD_ID_SETTLER_MOVE;
    this.name = "移動";
    this.cost = function() {
        return SETTLER_MOVE_COST;
    }
    this.localExecute = function() {
        if(this.toloc) {
            // 他人の星には移動できない
            var p = app.spaceMap[this.toloc];
            if ((p) && (p.ctl) && (p.ctl.id != this.empire.id)) {
                printConsole(this.empire.id, "移民船団移動失敗(他の勢力の植民星) : " + this.toloc);
                return false;
            }
            // 既にそこに移民船団がいるときには、そこへはいけない
            if (app.settlerMap[this.toloc] >= 0) {
                printConsole(this.empire.id, "移民船団移動失敗(別の船団がいる) : " + this.toloc);
                return false;
            }
            // 移動を実施
            this.empire.moveSettler(this.fromloc, this.toloc);
            printConsole(this.empire.id, "移民船団移動 : "
                            + this.fromloc + "->" + this.toloc);
            return true;
        }
        return false;
    }

    this.cmdExecute = cmdExecute;

    this.think = function() {
        // 一番近い空き地か人口の少ない星に移動する。
        var min = 999;
        var minlk = null;
        var undplList = this.empire.ai.findUndplList();
        if (undplList) {
            for (var lk in undplList) {
                var p = app.starList[lk];
                var d = calDistance(this.fromloc, p.loc);
                if (d < min) {
                    min = d;
                    minlk = p.loc;
                }
            }
        }
        
        for (var i = 0; i < this.empire.ctlPlanetList.length; i++) {
            var p = this.empire.ctlPlanetList[i];
            if (p.dpl > 3) {
                continue;
            }
            var d = calDistance(this.fromloc, p.loc);
            if (d < min) {
                min = d;
                minlk = p.loc;
            }
        }
        if (minlk) {
            this.moveTo(minlk);
            return true;
        }
        
        // それがなければ、一番近い未探査地点へ移動する。
        for (var d = this.empire.nvg + 1; true; d++) {
            var list = app.getLocListFromDistance(this.fromloc, d);
            if (list.length <= 0) {
                break;
            }
            list = list.randomize();
            for (var i = 0; i < list.length; i++) {
                if (! this.empire.visibleList[list[i]]) {
                    this.moveTo(list[i]);
                    return true;
                }
            }
        }
        return false;
    }

    this.moveTo = function(toloc) {
        for (var d = this.empire.nvg; d > 0; d--) {
            this.toloc = moveTo(this.fromloc, toloc, d);
            if (! this.empire.isVisible(this.toloc)) {
                continue;
            }
            if (app.spaceMap[this.toloc]) {
                continue;
            }
            if (app.uchuju.isUchujuLoc(this.toloc)) {
                continue;
            }
            return;
        }
        this.toloc = this.fromloc;
        return;
    }

    this.canBuild = function() {
        // 移動は常に可能
        return true;
    }
    this.canBuildList = function() {
        if (this.canBuild()) {
            return new Array(this);
        }
    }

    this.getFullName = function() {
        if (this.toloc) {
            return this.name + "(" + this.toloc + ")";
        } else {
            return this.name;
        }
    }

    this.makeKey = function() {
        return this.id;
    }
}

// 移民船団の移民
function CmdSettlerJoin(empire, fromloc, toloc) {
    this.empire = empire;
    this.fromloc = fromloc;
    this.toloc = (toloc) ? toloc : null;
    this.id = CMD_ID_SETTLER_JOIN;
    this.name = "移民";
    this.cost = function() {
        return SETTLER_JOIN_COST;
    }
    this.localExecute = function() {
        if(!this.toloc) {
            return false;
        }
        // 移民は既に支配している星だけ
        var p = app.starList[this.toloc];
        if (p) {
            if ((! p.ctl) || (p.ctl.id != this.empire.id)) {
                printConsole(this.empire.id, "移民船団移民失敗(他の勢力の植民星) : "
                        + this.toloc);
                return false;
            }
        } else {
            printConsole(this.empire.id, "移民船団移民失敗(そこに星はない) : " + this.toloc);
            return false;
        }
        // 移民を実施
        this.empire.removeSettler(this.fromloc);
        p.dpl++;
        printConsole(this.empire.id, "移民船団移民 : " + p.name);
        return true;
    }

    this.cmdExecute = cmdExecute;

    this.think = function() {
        var list = this.canBuildList(this.empire.ai, this.fromloc).randomize();
        for (var i = 0; i < list.length; i++) {
            var p = app.starList[list[i].toloc];
            if (p.dpl <= 3) {
                this.toloc = p.loc;
                return true;
            }
        }
        return false;
    }

    this.canBuild = function() {
        return (this.canBuildList(this.empire.ai, this.fromloc).length > 0);
    }

    // 移民可能なリストを作る
    this.canBuildList = function(ai, sloc) {
        var result = new Array();
        if (ai.empire.tmp_trs >= SETTLER_JOIN_COST) {
            var list = new Array();
            for (var i = 0; i < ai.empire.ctlPlanetList.length; i++) {
                var p = ai.empire.ctlPlanetList[i];
                if (p.dpl >= 9) {
                    continue;
                }
                var d = calDistance(sloc, p.loc);
                if (d <= this.empire.nvg) {
                    list.push(p);
                }
            }
            for (var i = 0; i < list.length; i++) {
                var loc = list[i].loc;
                if (app.uchuju.isUchujuLoc(loc) ) {
                    continue;
                }
                this.toloc = loc;
                result.push(clone(this));
            }
        }
        return result;
    }

    this.getFullName = function() {
        if (this.toloc && app.starList[this.toloc]) {
            return this.name + "(" + app.starList[this.toloc].name + ")";
        }
        return this.name;
    }

    this.makeKey = function() {
        return this.id + "-" + this.toloc;
    }
}

// 移民船団の植民
function CmdSettlerColony(empire, fromloc, toloc) {
    this.empire = empire;
    this.fromloc = fromloc;
    this.toloc = (toloc) ? toloc : null;
    this.id = CMD_ID_SETTLER_COLONY;
    this.name = "植民星建設";
    this.cost = function() {
        return SETTLER_COLONY_COST;
    }
    this.localExecute = function() {
        if(! this.toloc) {
            return false;
        }
        // 植民は所有者のいない星だけ
        var p = app.starList[this.toloc];
        if (p) {
            if (p.ctl) {
                printConsole(this.empire.id, "移民船団植民失敗(他の勢力の植民星) : "
                        + this.toloc);
                        return false;
            }
        } else {
            printConsole(this.empire.id, "移民船団植民失敗(そこに星はない) : " + this.toloc);
            return false;
        }
        // 植民を実施
        p.dpl = 1;
        p.changeCtl(this.empire);
        printConsole(this.empire.id, this.empire.name + "は"
                + p.name + "にコロニーを築きました");
        this.empire.removeSettler(this.fromloc);
        return true;
    }

    this.cmdExecute = cmdExecute;

    this.think = function() {
        return this.canBuild();
    }

    this.canBuild = function() {
        return (this.canBuildList(this.empire.ai, this.fromloc).length > 0);
    }

    // 植民可能なリストを作る
    this.canBuildList = function(ai, sloc) {
        var result = new Array();
        if (ai.empire.tmp_trs >= SETTLER_COLONY_COST) {
            var undplList = ai.findUndplListFromLoc(sloc, ai.empire.fnd);
            var list = ai.findColonizableList(undplList, ai.empire.nvg);
            for (var i = 0; i < list.length; i++) {
                var loc = list[i].loc;
                if (app.uchuju.isUchujuLoc(loc) ) {
                    continue;
                }
                this.toloc = loc;
                result.push(clone(this));
            }
        }
        return result;
    }

    this.getFullName = function() {
        if ( (this.toloc) && (app.starList[this.toloc]) ) {
            return this.name + "(" + app.starList[this.toloc].name + ")";
        }
        return this.name;
    }

    this.makeKey = function() {
        return this.id + "-" + this.toloc;
    }
}

// IDからインスタンスを返す
function getCmdById(id, empire, planet, target) {
    if (id == CMD_ID_ECN_UP) {
        return new CmdBuildEcn(empire, planet);
    } else if (id == CMD_ID_COLONY) {
        return new CmdBuildColony(empire, planet, target);
    } else if (id == CMD_ID_POP_MOVE) {
        return new CmdImin(empire, planet, target);
    } else if (id == CMD_ID_BASE) {
        return new CmdBuildDfc(empire, planet);
    } else if (id == CMD_ID_FLEET) {
        return new CmdBuildFleet(empire, planet);
    } else if (id == CMD_ID_SETTLER) {
        return new CmdBuildSettler(empire, planet);
    }
    return new CmdNothing();
}

// IDから移民船団コマンドのインスタンスを返す
function getSettlerCmdById(id, empire, fromloc, toloc) {
    if (id == CMD_ID_SETTLER_MOVE) {
        return new CmdSettlerMove(empire, fromloc, toloc);
    } else if (id == CMD_ID_SETTLER_JOIN) {
        return new CmdSettlerJoin(empire, fromloc, toloc);
    } else if (id == CMD_ID_SETTLER_COLONY) {
        return new CmdSettlerColony(empire, fromloc, toloc);
    }
    return new CmdNothing();
}

function clone(obj) {
    var result = new Object();
    for (var e in obj) {
        result[e] = obj[e];
    }

    return result;
}                                                                                                   